/*
 * Copyright (c) 2012 - 2013 Alexandre Tuleu
 *
 * This file is part of biorob-cpp.
 *
 *   biorob-cpp is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as
 *   published by the Free Software Foundation, either version 3 of the
 *   License, or (at your option) any later version.
 *
 *   biorob-cpp is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of both the GNU General Public
 *   License AND GNU Lesser General Public License along with
 *   biorob-cpp.  If not, see <http://www.gnu.org/licenses/>.
 */


/*
 * Copyright (c) 2012  Alexandre Tuleu
 *
 * This file is part of biorob-cpp.
 *
 *   biorob-cpp is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as
 *   published by the Free Software Foundation, either version 3 of the
 *   License, or (at your option) any later version.
 *
 *   biorob-cpp is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of both the GNU General Public
 *   License AND GNU Lesser General Public License along with
 *   PROJECT.  If not, see <http://www.gnu.org/licenses/>.
 */


/**
 * \file Command.h
 *
 *  \date Mar 27, 2012
 *  \author Alexandre Tuleu
 */

#ifndef BIOROB_CLI_COMMAND_H_
#define BIOROB_CLI_COMMAND_H_

#include <string>
#include <tr1/memory>
#include <vector>
#include <stdexcept>
#include <iostream>

#include "Variable.h"


namespace cli {
/**
 *
 */
class Command {
public:
	typedef std::tr1::shared_ptr<Command> Ptr;
	typedef std::vector<VariableBase::Ptr> ListOfArguments;

	Command(const std::string & name,
	        unsigned int nbArgs,
	        bool acceptOptionnalArgs = false);
	virtual ~Command();

	const std::string & Name() const;
	unsigned int NumberbOfArguments() const;
	bool AcceptOptionnalArgs() const;


	virtual bool Call(const ListOfArguments & args) = 0;
private :
	const std::string d_name;
	const unsigned int d_nbArgs;
	const bool d_acceptOptionnalArgs;
};


class KeyCallback {
public :
	typedef std::tr1::shared_ptr<KeyCallback> Ptr;

	KeyCallback(){}
	virtual ~KeyCallback(){}

	virtual int Call(int key) = 0;
};

template <typename T>
static T ParseArg(VariableBase::Ptr v);

template <class Base >
class ClosureBase : public Command {
public :
	ClosureBase(const std::string & name,
	            unsigned int nbArgs,
	            Base & base);
	virtual ~ClosureBase();

protected :
	Base & d_base;

};

template <class Base >
class Closure0 : public ClosureBase<Base> {
public :
	typedef void (Base::*MethodPtr)();
	Closure0(const std::string & name ,
	         Base & base ,
	         MethodPtr method);
	virtual ~Closure0();

	static Command::Ptr Create(const std::string & name ,
	                           Base & base ,
	                           MethodPtr method);

	bool Call(const Command::ListOfArguments & args);
private :
	MethodPtr d_method;
};

template <class Base, typename Arg1>
class Closure1 : public ClosureBase<Base> {
public :
	typedef void (Base::*MethodPtr)(const Arg1 &);
	Closure1(const std::string & name ,
	         Base & base,
	         MethodPtr method);
	virtual ~Closure1();

	static Command::Ptr Create(const std::string & name ,
	                           Base & base ,
	                           MethodPtr method);

	bool Call(const Command::ListOfArguments & args);
private :
	MethodPtr d_method;
};

template <class Base, typename Arg1, typename Arg2>
class Closure2 : public ClosureBase<Base> {
public :
	typedef void (Base::*MethodPtr)(const Arg1 &, const Arg2 & );
	Closure2(const std::string & name ,
	         Base & base,
	         MethodPtr method);
	virtual ~Closure2();

	static Command::Ptr Create(const std::string & name ,
	                           Base & base ,
	                           MethodPtr method);

	bool Call(const Command::ListOfArguments & args);
private :
	MethodPtr d_method;
};

/*******************************************************************************
 * INLINE IMPLEMENTATION
 *****************************************************************************/


template <class Base>
inline ClosureBase<Base>::ClosureBase(const std::string & name,
                                      unsigned int nbArgs,
                                      Base & base)
	: Command(name,nbArgs,false)
	, d_base(base){
}

template <class Base>
inline ClosureBase<Base>::~ClosureBase(){}

template <typename T>
inline T ParseArg(VariableBase::Ptr var){
	std::istringstream iss(var->toString());

	T val;
	iss >> val;
	if(iss.fail()){
		throw std::runtime_error("could not parse '" + var->toString() + "'");
	}
	return val;
}


template <class Base>
inline Closure0<Base>::Closure0(const std::string & name ,
                                Base & base ,
                                MethodPtr method)
	: ClosureBase<Base>(name,0,base)
	, d_method(method){
}

template <class Base>
inline Closure0<Base>::~Closure0(){}

template <class Base>
bool Closure0<Base>::Call(const Command::ListOfArguments & ){
	try {
		(ClosureBase<Base>::d_base.*d_method)();
	} catch (std::runtime_error & e) {
		std::cerr << "<<< " << e.what() << std::endl;
		return false;
	}
	return true;
}

template <class Base>
inline Command::Ptr Closure0<Base>::Create(const std::string & name ,
                                    Base & base ,
                                    MethodPtr method) {
	return Command::Ptr(new Closure0<Base>(name,base,method));
}


template <class Base,typename Arg1>
inline Closure1<Base,Arg1>::Closure1(const std::string & name ,
                                     Base & base ,
                                     MethodPtr method)
	: ClosureBase<Base>(name,1,base)
	, d_method(method){
}

template <class Base,typename Arg1>
inline Closure1<Base,Arg1>::~Closure1(){}


template <class Base,typename Arg1>
inline bool Closure1<Base,Arg1>::Call(const Command::ListOfArguments & args){
	try {
		Arg1 a1 = ParseArg<Arg1>(args.front());
		(ClosureBase<Base>::d_base.*d_method)(a1);
	} catch (std::runtime_error & e) {
		std::cerr << "<<< " << e.what() << std::endl;
		return false;
	}
	return true;
}

template <class Base,typename Arg1>
inline Command::Ptr Closure1<Base,Arg1>::Create(const std::string & name ,
                                    Base & base ,
                                    MethodPtr method) {
	return Command::Ptr(new Closure1<Base,Arg1>(name,base,method));
}


template <class Base,typename Arg1,typename Arg2>
inline Closure2<Base,Arg1,Arg2>::Closure2(const std::string & name ,
                                          Base & base ,
                                          MethodPtr method)
	: ClosureBase<Base>(name,2,base)
	, d_method(method){
}

template <class Base,typename Arg1,typename Arg2>
inline Closure2<Base,Arg1,Arg2>::~Closure2(){}


template <class Base,typename Arg1,typename Arg2>
inline bool Closure2<Base,Arg1,Arg2>::Call(const Command::ListOfArguments & args){
	try {
		Arg1 a1 = ParseArg<Arg1>(args[0]);
		Arg2 a2 = ParseArg<Arg2>(args[1]);
		(ClosureBase<Base>::d_base.*d_method)(a1,a2);
	} catch (std::runtime_error & e) {
		std::cerr << "<<< " << e.what() << std::endl;
		return false;
	}
	return true;
}

template <class Base,typename Arg1,typename Arg2>
inline Command::Ptr Closure2<Base,Arg1,Arg2>::Create(const std::string & name ,
                                                     Base & base ,
                                                     MethodPtr method) {
	return Command::Ptr(new Closure2<Base,Arg1,Arg2>(name,base,method));
}

}

#endif // BIOROB_CPP_CLI_COMMAND_H_
