/*
 * Copyright (c) 2012 - 2013 Alexandre Tuleu
 *
 * This file is part of biorob-cpp.
 *
 *   biorob-cpp is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as
 *   published by the Free Software Foundation, either version 3 of the
 *   License, or (at your option) any later version.
 *
 *   biorob-cpp is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of both the GNU General Public
 *   License AND GNU Lesser General Public License along with
 *   biorob-cpp.  If not, see <http://www.gnu.org/licenses/>.
 */


/*
 * Copyright (c) 2012  Alexandre Tuleu
 *
 * This file is part of biorob-cpp.
 *
 *   biorob-cpp is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as
 *   published by the Free Software Foundation, either version 3 of the
 *   License, or (at your option) any later version.
 *
 *   biorob-cpp is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of both the GNU General Public
 *   License AND GNU Lesser General Public License along with
 *   PROJECT.  If not, see <http://www.gnu.org/licenses/>.
 */


/**
 * \file Option.h
 *
 *  \date Nov 28, 2011
 *  \author Alexandre Tuleu
 */

#ifndef BIOROB_CPP_OPTIONS_OPTION_H_
#define BIOROB_CPP_OPTIONS_OPTION_H_

#include <map>
#include <vector>
#include <tr1/memory>
#include <string>
#include <iostream>
#include <sstream>
#include <algorithm>

namespace options {

	class SectionAndOptionBase {
	public :
		SectionAndOptionBase(const std::string & name,
		                  const std::string & comment);
		virtual ~SectionAndOptionBase();

		const std::string & Name() const;
		const std::string & Comment() const;
		char Short() const;

		void SetComment(const std::string & comment);

		bool IsUnknown() const;
		void SetIsUnknown(bool isUnknown=true);

	private :
		void SetNameFromString(const std::string & name);


		std::string d_name,d_comment;
		char d_short;
		bool d_isUnknown;

	};

	template <typename T>
	class Option;

	template <typename T>
	class OptionList;

	class OptionBase : public SectionAndOptionBase {
	public :
		typedef std::tr1::shared_ptr<OptionBase> Ptr;

		OptionBase(const std::string & name,
		           const std::string & comment = std::string());

		virtual ~OptionBase();

		bool IsDefault() const;

		template <typename T>
		bool GetValue(T & value) const;

		virtual bool Parse (const std::string & value) = 0;

		virtual const std::string & ValueAsString() const = 0;


	protected :
		virtual std::auto_ptr<OptionBase> Clone() const = 0;

		void SetIsDefault(bool def=true);


	private :

		bool        d_isDefault;
		bool        d_isUnknown;
	};


	class Section : public SectionAndOptionBase{
	public :
		typedef std::tr1::shared_ptr<Section> Ptr;
		class SubSectionCreator {
		public :
			virtual ~SubSectionCreator();
			typedef std::tr1::shared_ptr<SubSectionCreator> Ptr;
			virtual Section::Ptr Create(const std::string & name,
			                            const std::string & comment);
		};
		typedef std::map<std::string,Ptr>     List;
		typedef std::map<std::string,OptionBase::Ptr> ListOfOptions;

		Section(const std::string & name, const std::string & comment);
		virtual ~Section();

		void SetSubSectionCreator(SubSectionCreator::Ptr ssc);

		const List & SubSections() const;
		List & SubSections();

		const Ptr SubSection(const std::string & name) const;
		Ptr SubSection(const std::string & name);

		Ptr AddSubSection(const std::string & name,
		                  const std::string & comment);

		const ListOfOptions & Options() const;
		ListOfOptions & Options();

		const OptionBase::Ptr GetOption(const std::string & name) const;
		OptionBase::Ptr GetOption(const std::string & name);


		template <typename T>
		std::tr1::shared_ptr<Option <T> >
		AddOption(const std::string & name,
		          const T & defaultValue,
		          const std::string & comment);

		template <typename T>
		std::tr1::shared_ptr<OptionList <T> >
		AddOptionList(const std::string & name,
		              const std::string & comment);




		Ptr DeepCopy(const std::string & newName) const;

	protected :
		Ptr DeepCopy() const;

	private :
		List d_subSections;
		ListOfOptions d_options;
		SubSectionCreator::Ptr d_subSectionCreator;
	};

	template <typename T>
	class Option : public OptionBase {
	public:
		typedef std::tr1::shared_ptr<Option <T> > Ptr;

		Option(const std::string & name,
		       const T & defaultValue,
	           const std::string & comment = std::string());
		virtual ~Option();

		virtual bool Parse (const std::string & value);

		virtual const std::string & ValueAsString() const;

		const T & Value() const;

		void SetValue(const T & value );

	protected :
		virtual std::auto_ptr<OptionBase> Clone() const;

	private :
		void PrintValueInString();

		T           d_value;
		std::string d_valueStr;
	};

	template <>
	class Option<std::string> : public OptionBase {
	public:
		typedef std::tr1::shared_ptr<options::Option<std::string> > Ptr;


		Option(const std::string & name,
		       const std::string & defaultValue,
	           const std::string & comment = std::string());
		virtual ~Option();

		virtual bool Parse (const std::string & value);

		virtual const std::string & ValueAsString() const;

		const std::string & Value() const;

		void SetValue(const std::string & value);

	protected :
		virtual std::auto_ptr<OptionBase> Clone() const;

	private :
		std::string d_value;
	};

	template <>
	class Option<bool> : public OptionBase {
	public:
		typedef std::tr1::shared_ptr<options::Option<bool> > Ptr;

		Option(const std::string & name,
		       const bool defaultValue,
	           const std::string & comment = std::string());
		virtual ~Option();

		virtual bool Parse (const std::string & value);

		virtual const std::string & ValueAsString() const;

		bool Value() const;

		void SetValue(bool value);

	protected :
		virtual std::auto_ptr<OptionBase> Clone() const;

	private :
		const static std::string s_true,s_false;

		bool        d_value;
	};

	template <typename T>
	class OptionList : public OptionBase {
	public :
		typedef std::vector<T> List;
		OptionList(const std::string & name,
		           const std::string & comment = std::string());
		virtual ~OptionList();

		virtual bool Parse (const std::string & value);

		virtual const std::string & ValueAsString() const;

		const List & Values() const;

	protected :
		virtual std::auto_ptr<OptionBase> Clone() const;

	private :
		std::vector<T>     d_values;
		std::string        d_valuesStr;

	};


	template <typename T>
	inline std::tr1::shared_ptr<Option<T> >
	Section::AddOption(const std::string & name,
	                   const T & defaultValue,
	                   const std::string & comment) {
		typedef std::tr1::shared_ptr<Option<T> > OptPtr;
		if(name.empty()){
			return OptPtr();
		}
		OptPtr so(new Option<T>(name,defaultValue,comment));

		d_options[so->Name()] = std::tr1::static_pointer_cast<OptionBase,Option<T> >(so);

		return so;
	}

	template <typename T>
	inline std::tr1::shared_ptr<OptionList<T> >
	Section::AddOptionList(const std::string & name,
	                                             const std::string & comment){
		if(name.empty()){
			return std::tr1::shared_ptr<OptionList<T> >();
		}
		std::tr1::shared_ptr<OptionList<T> > so(new OptionList<T>(name,comment));

		d_options[so->Name()] = std::tr1::static_pointer_cast<OptionBase,OptionList<T> >(so);

		return so;
	}

	template <typename T>
	inline bool OptionBase::GetValue(T & value) const{
		std::istringstream iss(this->ValueAsString());

		iss >> value;

		return (iss.good() || iss.eof());
	}




	template <typename T>
	inline Option<T>::Option(const std::string & name,
	                  const T & defaultValue,
	                  const std::string & comment)
		: OptionBase(name,comment)
		, d_value(defaultValue)
		, d_valueStr(){
		PrintValueInString();
	}

	template <typename T>
	inline Option<T>::~Option(){}


	template <typename T>
	inline bool Option<T>::Parse(const std::string & value){
		std::istringstream iss(value);
		iss >> d_value;
		if(!iss.fail()){
			SetIsDefault(false);
			PrintValueInString();
			return true;
		}
		return false;
	}

	template <typename T>
	inline const std::string & Option<T>::ValueAsString() const{
		return d_valueStr;
	}

	template <typename T>
	inline const T & Option<T>::Value() const {
		return d_value;
	}



	template <typename T>
	inline void Option<T>::SetValue(const T & value){
		d_value = value;
		PrintValueInString();
	}



	template <typename T>
	inline void Option<T>::PrintValueInString(){
		std::ostringstream os;
		os << d_value << std::flush ;
		d_valueStr = os.str();
	}

	template <typename T>
	inline std::auto_ptr<OptionBase> Option<T>::Clone() const {
		return std::auto_ptr<OptionBase>(new Option<T>(*this));
	}

	inline Option<bool>::Option(const std::string & name,
	                            bool defaultValue,
	                            const std::string & comment)
		: OptionBase(name,comment)
		, d_value(defaultValue){
	}


	inline Option<bool>::~Option(){}


	inline bool Option<bool>::Parse(const std::string & value){
		std::string val = value;
		//remove all white spaces.
		val.erase(remove_if(val.begin(), val.end(), isspace), val.end());

		if(val.empty()
		   || val == "no"
		   || val == "false"
		   || val == "0"
		   || val == "off"){
			d_value = false;
		} else {
			d_value = true;
		}
		SetIsDefault(false);

		return true;
	}


	inline const std::string & Option<bool>::ValueAsString() const{
		return d_value ? s_true : s_false;
	}


	inline  bool Option<bool>::Value() const {
		return d_value;
	}

	inline  void Option<bool>::SetValue(bool value){
		d_value = value;
	}


	inline std::auto_ptr<OptionBase> Option<bool>::Clone() const {
		return std::auto_ptr<OptionBase> (new Option<bool>(*this));
	}

	inline Option<std::string>::Option(const std::string & name,
	                                   const std::string & defaultValue,
	                                   const std::string & comment)
		: OptionBase(name,comment)
		, d_value(defaultValue){
	}

	inline Option<std::string>::~Option() {}


	inline bool  Option<std::string>::Parse (const std::string & value) {
		d_value = value;
		SetIsDefault(false);
		return true;
	}

	inline const std::string & Option<std::string>::ValueAsString() const {
		return d_value;
	}

	inline const std::string & Option<std::string>::Value() const {
		return d_value;
	}



	inline void Option<std::string>::SetValue(const std::string & value){
		d_value = value;
	}


	inline std::auto_ptr<OptionBase> Option<std::string>::Clone() const {
		return std::auto_ptr<OptionBase> (new Option<std::string>(*this));
	}


	template <typename T>
	inline OptionList<T>::OptionList(const std::string & name,
	                                 const std::string & comment)
		: OptionBase(name,comment)
		, d_values()
		, d_valuesStr(){

	}

	template <typename T>
	inline OptionList<T>::~OptionList(){}


	template <typename T>
	inline bool OptionList<T>::Parse (const std::string & values){
		std::istringstream iss(values);
		std::string v;
		while(std::getline(iss,v,';')){
			T val;
			std::istringstream iss2(v);
			iss2 >> val;
			if (!iss2.fail()){
				d_values.push_back(val);
				if(!d_valuesStr.empty()){
					d_valuesStr.append(";");
				}
				d_valuesStr.append(v);
			}
		}
		SetIsDefault(false);
		return true;
	}

	template <typename T>
	inline const std::string & OptionList<T>::ValueAsString() const{
		return d_valuesStr;
	}

	template <typename T>
	inline const typename OptionList<T>::List & OptionList<T>::Values() const{
		return d_values;
	}

	template <typename T>
	inline std::auto_ptr<OptionBase> OptionList<T>::Clone() const {
		return std::auto_ptr<OptionBase>(new OptionList<T>(*this));
	}

}
#endif // BIOROB_CPP_OPTIONS_OPTION_H_
